# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"query_API": "01-retrieval.ipynb",
         "dict_col_2_cols": "01-retrieval.ipynb",
         "clean_nested_dict_cols": "01-retrieval.ipynb",
         "set_dt_idx": "01-retrieval.ipynb",
         "create_df_dt_rng": "01-retrieval.ipynb",
         "clean_df_dts": "01-retrieval.ipynb",
         "retrieve_stream_df": "01-retrieval.ipynb",
         "check_streams": "01-retrieval.ipynb",
         "retrieve_streams_df": "01-retrieval.ipynb",
         "parse_A44_response": "01-retrieval.ipynb",
         "retreive_DAM_prices": "01-retrieval.ipynb",
         "parse_A75_response": "01-retrieval.ipynb",
         "retrieve_production": "01-retrieval.ipynb",
         "load_EI_df": "02-eda.ipynb",
         "load_DE_df": "02-eda.ipynb",
         "clean_df_for_plot": "02-eda.ipynb",
         "rgb_2_plt_tuple": "02-eda.ipynb",
         "convert_fuel_colour_dict_to_plt_tuple": "02-eda.ipynb",
         "hide_spines": "02-eda.ipynb",
         "stacked_fuel_plot": "02-eda.ipynb",
         "get_dist": "03-lowess.ipynb",
         "get_dist_threshold": "03-lowess.ipynb",
         "dist_to_weights": "03-lowess.ipynb",
         "get_all_weights": "03-lowess.ipynb",
         "vector_to_dist_matrix": "03-lowess.ipynb",
         "get_frac_idx": "03-lowess.ipynb",
         "get_dist_thresholds": "03-lowess.ipynb",
         "clean_weights": "03-lowess.ipynb",
         "dist_2_weights_matrix": "03-lowess.ipynb",
         "get_full_dataset_weights_matrix": "03-lowess.ipynb",
         "get_weighting_locs": "03-lowess.ipynb",
         "create_dist_matrix": "03-lowess.ipynb",
         "num_fits_2_reg_anchors": "03-lowess.ipynb",
         "get_weights_matrix": "03-lowess.ipynb",
         "calc_lin_reg_betas": "03-lowess.ipynb",
         "fit_regressions": "03-lowess.ipynb",
         "check_array": "03-lowess.ipynb",
         "lowess_fit_and_predict": "03-lowess.ipynb",
         "calc_robust_weights": "03-lowess.ipynb",
         "robust_lowess_fit_and_predict": "03-lowess.ipynb",
         "Lowess": "03-lowess.ipynb",
         "get_bootstrap_idxs": "03-lowess.ipynb",
         "get_bootstrap_resid_std_devs": "03-lowess.ipynb",
         "run_model": "03-lowess.ipynb",
         "bootstrap_model": "03-lowess.ipynb",
         "get_confidence_interval": "03-lowess.ipynb",
         "pred_to_quantile_loss": "03-lowess.ipynb",
         "calc_quant_reg_loss": "03-lowess.ipynb",
         "calc_quant_reg_betas": "03-lowess.ipynb",
         "quantile_model": "03-lowess.ipynb",
         "calc_timedelta_dists": "03-lowess.ipynb",
         "construct_dt_weights": "03-lowess.ipynb",
         "fit_external_weighted_ensemble": "03-lowess.ipynb",
         "get_ensemble_preds": "03-lowess.ipynb",
         "process_smooth_dates_fit_inputs": "03-lowess.ipynb",
         "SmoothDates": "03-lowess.ipynb",
         "construct_pred_ts": "05-price-moe.ipynb",
         "LowessDates": "03-lowess.ipynb",
         "PicklableFunction": "04-price-surface-estimation.ipynb",
         "get_fit_kwarg_sets": "04-price-surface-estimation.ipynb",
         "fit_models": "04-price-surface-estimation.ipynb",
         "construct_dispatchable_lims_df": "05-price-moe.ipynb",
         "construct_pred_mask_df": "05-price-moe.ipynb",
         "AxTransformer": "05-price-moe.ipynb",
         "set_ticks": "05-price-moe.ipynb",
         "set_date_ticks": "05-price-moe.ipynb",
         "construct_df_pred": "05-price-moe.ipynb",
         "calc_error_metrics": "05-price-moe.ipynb",
         "get_model_pred_ts": "05-price-moe.ipynb",
         "weighted_mean_s": "05-price-moe.ipynb",
         "app": "10-ci-cd.ipynb",
         "get_current_package_version": "10-ci-cd.ipynb",
         "increment_package_version": "10-ci-cd.ipynb",
         "set_current_package_version": "10-ci-cd.ipynb"}

modules = ["retrieval.py",
           "eda.py",
           "lowess.py",
           "surface.py",
           "moe.py",
           "cicd.py"]

doc_url = "https://AyrtonB.github.io/Merit-Order-Effect/"

git_url = "https://github.com/AyrtonB/Merit-Order-Effect"

def custom_doc_links(name): return None
